""" Implementation of mooda.read_pkl(path) """
import pickle
from .. import WaterFrame

def read_pkl(path_pkl):
    """
    Get a WaterFrame from a pickle file.

    Parameters
    ----------
        path_pkl: str
            Location of the pickle file.

    Returns
    -------
        wf_pkl: WaterFrame
    """
    wf_pkl = WaterFrame()

    pickle_dataset = pickle.load(open(path_pkl, "rb"))

    wf_pkl.data = pickle_dataset.get('data')
    wf_pkl.vocabulary = pickle_dataset.get('vocabulary')
    wf_pkl.metadata = pickle_dataset.get('metadata')

    return wf_pkl



""" Implementation of mooda """
from .waterframe import WaterFrame
from .input import read_nc_emodnet, read_nc, read_nc_imos, read_pkl, read_nc_moist
from .util import concat, iplot_location, iplot_timeseries, md5, es_create_indexes

__version__ = '1.4.4'



""" Implementation of a function to read a generic NetCDF file """
import xarray as xr
from ..waterframe import WaterFrame

def read_nc(path, decode_times=True):
    """
    Read data form NetCDF file and create a WaterFrame.

    Parameters
    ----------
        path: str
            Path of the NetCDF file.
        decode_times : bool, optional
            If True, decode times encoded in the standard NetCDF datetime format 
            into datetime objects. Otherwise, leave them encoded as numbers. 
    
    Returns
    -------
        wf: WaterFrame
    """
    # Create WaterFrame
    wf = WaterFrame()

    # Open file with xarrat
    ds = xr.open_dataset(path, decode_times=decode_times)

    # Save ds into a WaterFrame
    wf.metadata = dict(ds.attrs)
    wf.data = ds.to_dataframe()
    for variable in ds.variables:
        wf.vocabulary[variable] = dict(ds[variable].attrs)

    return wf



""" Implementation of WaterFrame.rename(actual_name, new_name) """


def rename(self, actual_name, new_name, inplace=True):
    """
    It renames a parameter.

    Parameters
    ----------
        actual_name: str
            Actual name of a parameter.
        new_name: str
            New name of the parameter.
        inplace: bool
            If True, the rename is inplace and returns True

    Returns
    -------
        new_wf: WaterFrame
            WaterFrame with the renamed parameters.
    """

    new_data = self.data.rename(
        columns={actual_name: new_name, f'{actual_name}_QC': f'{new_name}_QC'},
        inplace=inplace)

    if inplace:
        new_wf = True
    else:
        new_wf = self.copy()
        new_wf.data = new_data

    return new_wf



""" Implementation of a function to read NetCDF files from EMODNET """
import xarray as xr
from ..waterframe import WaterFrame


def read_nc_emodnet(path, clean_data=True):
    """
    Open a NetCDF file from EMODNET physics. The file should be from a fixed point observatory due
    to this function delete the coordenates LATITUDE and LONGITUDE from the variables. If LATITUDE
    and LONGITUDE are not deleted, the pandas DataFrame is huge and your script will raise a
    MemoryError.

    Parameters
    ----------
        path: str
            Path of the NetCDF file.
        clean_data: bool
            It deletes variables with DC in name.

    Returns
    -------
        wf: WaterFrame
    """
    def drop(ds_in, clean_data_in):
        """Drop some parameters of the dataset.

        Parameters
        ----------
            ds_in: xarray.Dataset
                Input dataset.
            clean_data_in: bool
                It indicates if "_DM" variables should be deleted.

        Returns
        -------
            ds_out: xarray.Dataset
                Output dataset"""
        # Creation of a list with the variables to drop. We are going to
        # drop all variables with a "_DM" in the key name.
        vars2drop = [k for k in ds_in.variables.keys() if '_DM' in k]

        if clean_data_in:
            # Dropping the previous list and some other variables
            ds_out = ds_in.drop(vars2drop)
        else:
            ds_out = ds_in

        # ds_out = ds_out.drop_sel('DEPTH')
        if 'LATITUDE' in ds_in.variables.keys():
            ds_out = ds_out.drop('LATITUDE')
        if 'LONGITUDE' in ds_in.variables.keys():
            ds_out = ds_out.drop('LONGITUDE')
        if 'POSITION_QC' in ds_in.variables.keys():
            ds_out = ds_out.drop('POSITION_QC')
        if 'DC_REFERENCE' in ds_in.variables.keys():
            ds_out = ds_out.drop('DC_REFERENCE')
        if 'POSITIONING_SYSTEM' in ds_in.variables.keys():
            ds_out = ds_out.drop('POSITIONING_SYSTEM')
        return ds_out

    # Create WaterFrame
    wf = WaterFrame()

    # Open file with xarrat
    ds = xr.open_dataset(path)
    ds = drop(ds, clean_data)

    # Save ds into a WaterFrame
    wf.metadata = dict(ds.attrs)
    # Add network
    wf.metadata['network'] = 'emodnet'

    wf.data = ds.to_dataframe()
    # Harmonize data
    wf.data.reset_index(inplace=True)
    wf.data.set_index(['DEPH', 'TIME'], inplace=True)
    wf.data.index.rename(['DEPTH', 'TIME'], inplace=True)
    del wf.data['DEPTH']
    wf.data.rename(columns={'DEPH_QC':'DEPTH_QC'}, inplace=True)

    for variable in ds.variables:
        wf.vocabulary[variable] = dict(ds[variable].attrs)

    return wf



""" Implementation of WaterFrame.concat(waterframe) """
import pandas as pd


def concat(list_wf):
    """
    The concat function does all of the heavy lifting of performing concatenation operations between
    a list of WaterFrames.

    Parameters
    ----------
        list_wf: List of WaterFrames
            List of WaterFrames to be concatenated.

    Returns
    -------
        wf_one: WaterFrame
            WaterFrame with input data, metadata, and vocabulary concatenated.
    """
    for index, wf in enumerate(list_wf):
        if index == 0:
            wf_one = wf.copy()
        else:
            # Check if waterframes are from the same platform
            if wf.metadata.get('platform_name') == wf_one.metadata.get('platform_name'):

                # Same metadata, just change the coverage times
                start_wf = wf.metadata.get('time_coverage_start')
                end_wf = wf.metadata.get('time_coverage_end')

                start_wf_one = wf_one.metadata.get('time_coverage_start')
                end_wf_one = wf_one.metadata.get('time_coverage_end')

                if start_wf < start_wf_one:
                    wf_one.metadata['time_coverage_start'] = start_wf
                if end_wf > end_wf_one:
                    wf_one.metadata['time_coverage_end'] = end_wf

                # Concat data
                wf_one.data = pd.concat([wf_one.data, wf.data])
            else:
                raise NotImplementedError("All WaterFrames must be from the same 'platform_name'")

    return wf_one



""" Implementation of WaterFrame.qc_spike_test(parameters=None, window=0, threshold=3, flag=4)"""
import numpy as np


def qc_spike_test(self, parameters=None, window=0, threshold=3.5, influence=0.5, flag=4,
                  inplace=True):
    """
    Based on:
    https://stackoverflow.com/questions/22583391/peak-signal-detection-in-realtime-timeseries-data

    Algorithm based on the principle of dispersion: if a new datapoint is a given x number of
    standard deviations away from some moving mean, the algorithm signals (also called z-score).
    The algorithm is very robust because it constructs a separate moving mean and deviation, such
    that signals do not corrupt the threshold. Future signals are therefore identified with
    approximately the same accuracy, regardless of the amount of previous signals.
    
    For example, a window of 5 will use the last 5 observations to smooth
    the data. A threshold of 3.5 will signal if a datapoint is 3.5 standard deviations away
    from the moving mean. And an influence of 0.5 gives signals half of the influence that normal
    datapoints have. Likewise, an influence of 0 ignores signals completely for recalculating the
    new threshold. An influence of 0 is therefore the most robust option (but assumes stationarity);
    putting the influence option at 1 is least robust. For non-stationary data, the influence
    option should therefore be put somewhere between 0 and 1.

    Parameters
    ----------
        parameters: string or list of strings, optional
        (parameters = None)
            key of self.data to apply the test.
        window: int, optional (window = 0)
            The lag of the moving window.
            Minimun value = 3.
            If it is 0, the window is the 10% of the length of data with a maximum value of 100.
        threshold: float, optional (threshold = 3.5)
            The z-score at which the algorithm signals.
        influence: float, optional (influence = 0.5)
            The influence (between 0 and 1) of new signals on the mean and standard deviation.
        flag: int, optional (flag = 4)
            Flag value to write on the signal values.
        inplace: bool
            If True, it changes the flags in place and returns True.
            Otherwhise it returns an other WaterFrame.

    Returns
    -------
        new_wf: WaterFrame.
    """

    def thresholding_algo(y, lag, threshold, influence, signals, flag):
        filteredY = np.array(y)
        avgFilter = [0]*len(y)
        stdFilter = [0]*len(y)
        avgFilter[lag - 1] = np.mean(y[0:lag])
        stdFilter[lag - 1] = np.std(y[0:lag])
        for i in range(lag, len(y)):
            if abs(y[i] - avgFilter[i-1]) > threshold * stdFilter [i-1]:
                signals[i] = flag

                filteredY[i] = influence * y[i] + (1 - influence) * filteredY[i-1]
                avgFilter[i] = np.mean(filteredY[(i-lag+1):i+1])
                stdFilter[i] = np.std(filteredY[(i-lag+1):i+1])
            else:
                filteredY[i] = y[i]
                avgFilter[i] = np.mean(filteredY[(i-lag+1):i+1])
                stdFilter[i] = np.std(filteredY[(i-lag+1):i+1])

        return np.asarray(signals)

    if parameters is None:
        parameters = self.parameters
    elif isinstance(parameters, str):
        parameters = [parameters]

    data = self.data.copy()

    for parameter in parameters:
        # Auto calculation of window
        if window == 0:
            window = round(len(data[parameter])*5/100)  # 5% of len
            if window < 3:
                window = 3
            elif window > 100:
                window = 100

        df = data[[parameter, f'{parameter}_QC']].reset_index()
        df.set_index('TIME', inplace=True)

        for depth, df_depth in df.groupby('DEPTH'):
            df_depth.sort_index(inplace=True)

            y = df_depth[parameter].values
            signals = df_depth[f'{parameter}_QC'].values

            # Run algo with settings from above
            result = thresholding_algo(y, lag=window, threshold=threshold,
                                       influence=influence, signals=signals,
                                       flag=4)
            
            data.loc[(depth,), f'{parameter}_QC'] = result
    
    if inplace:
        self.data = data
        return True
    else:
        new_wf = self.copy()
        new_wf.data = data
        return new_wf

    return True



""" Implementation of WaterFrame.info_vocabulary(keys=None) """


def info_vocabulary(self, keys=None):
    """
    It returns a formatted string with the vocabulary information.

    Parameters
    ----------
        keys: string or list of strings (optional)
            The return message will contain the information of the input keys.
            If keys is None, all keys will be added to the return message.

    Returns
    -------
        message: str
            Message with the vocabulary information.
    """

    if keys is None:
        keys = self.vocabulary.keys()

    message = ""
    for key, vocabulary_dict in self.vocabulary.items():
        if key in keys:
            message += "  - {}\n".format(key)
            for vocabulary_key, vocabulary_value in vocabulary_dict.items():
                if str(vocabulary_value).strip() != "":
                    message += "    - {}: {}\n".format(vocabulary_key, vocabulary_value)

    return message[:-1]



""" Implementation of WaterFrame.qc_range_test(parameters=None, flag=4, limits=None) """


def qc_range_test(self, parameters=None, limits=None, flag=4, inplace=True):
    """
    Check if the values of a parameter are out of range.

    Parameters
    ----------
        parameters: string or list of strings, optional
        (parameters = None)
            key of self.data to apply the test.
        flag: int, optional (flag = 4)
            Flag value to write in on the fail values.
        limits: tuple, optional (limits = None)
            (Min value, max value) of the range of correct values.
        inplace: bool
            If True, it changes the flags in place and returns True.
            Otherwhise it returns an other WaterFrame.

    Returns
    -------
        new_wf: WaterFrame
    """
    ranges = {
        'ATMP': (600, 1500),  # atmospheric pressure at altitude
        'ATMS': (0, 2000),  # Atmospheric pressure at sea level
        'CHLT': (0, 30),  # total chlorophyll
        'CNDC': (0, 30),  # electrical conductivity
        'DRYT': (-20, 60),  # air temperature at sea level
        'GSPD': (0, 40),  # gust wind speed
        'HCDT': (0, 360),  # current to direction relative true north
        'HEAD': (0, 360),  # PLAT. HEADING REL. TRUE NORTH
        'LINC': (0, 3000),  # long-wave incoming radiation
        'LW': (0, 50),  # Downwelling vector radiance as energy
        'OSAT': (0, 200),  # oxygen saturation
        'PHPH': (0, 14),  # ph
        'PRES': (0, 500),  # sea pressure
        'PRRT': (0, 200),  # hourly precipitation rate
        'PSAL': (0, 40),  # practical salinity
        'RDIN': (0, 1100),  # incident radiation
        'SVEL': (130, 180000),  # sound velocity
        'SWDR': (0, 360),  # SWELL DIRECTION REL TRUE N.
        'SWHT': (0, 50),  # SWELL HEIGHT
        'SWPR': (0, 20),  # SWELL PERIOD
        'TEMP': (0, 50),  # Sea temperature
        'VAVH': (0, 20),  # AVER. HEIGHT HIGHEST 1/3 WAVE
        'VAVT': (0, 20),  # AVER. PERIOD HIGHEST 1/3 WAVE
        'VCMX': (0, 20),  # MAX CREST TROUGH WAVE HEIGHT
        'VDIR': (0, 360),  # wave direction rel. true north
        'VEPK': (0, 100),  # WAVE SPECTRUM PEAK ENERGY
        'VHM0': (0, 20),  # SPECTRAL SIGNIFICANT WAVE HEIGHT
        'VMDR': (0, 360),  # Mean wave direction
        'VPED': (0, 360),  # dir. spreading at wave peak
        'VPSP': (0, 360),  # dir. spreading at wave peak
        'VSMC': (0, 20),  # SPECTUM MOMENT(0, 2) WAVE PERIOD
        'VTDH': (0, 40),  # significant wave height
        'VTM02': (0, 13),  # Spectral moments (0, 2) wave period (Tm02)
        'VTPK': (0, 40),  # WAVE SPECTRUM PEAK PERIOD
        'VTZA': (0, 40),  # AVER ZERO CROSSING WAVE PERIOD
        'VTZM': (0, 40),  # period of the highest wave
        'VZMX': (0, 20),  # Maximum zero crossing wave height
        'WDIR': (0, 360),  # Wind from direction relative true north
        'WSPD': (0, 50),  # Horizontal wind speed
    }

    if parameters is None:
        parameters = self.parameters
    elif isinstance(parameters, str):
        parameters = [parameters]

    data = self.data.copy()

    for parameter in parameters:
        if limits:
            data.loc[self.data[parameter] < limits[0], parameter + '_QC'] = flag
            data.loc[self.data[parameter] > limits[1], parameter + '_QC'] = flag
        elif parameter in ranges.keys():
            self.data.loc[self.data[parameter] < ranges[parameter][0], parameter + '_QC'] = flag
            self.data.loc[self.data[parameter] > ranges[parameter][1], parameter + '_QC'] = flag
    
    if inplace:
        self.data = data
        return True
    else:
        new_wf = self.copy()
        new_wf.data = data
        return new_wf



""" Implementation of WaterFrame.data_intervals(parameter) """


def time_intervals(self, parameter, frequency):
    """
    It returns the index (TIME) of intervals between NaNs.

    Parameters
    ----------
        parameter: str
            Column name of WaterFrame.data.
        frequency: DateOffset, Timedelta or str
            Theorical sample frequency.

    Returns
    -------
        intervals: [(str, str)]
            List of tuples with the start and end index (TIME) of each interval of
            data.
    """
    # Creation of the DataFrame with the required info
    df = self.data.copy()

    df.reset_index(inplace=True)
    df.set_index('TIME', inplace=True)
    df.sort_index(inplace=True)

    # print(df[parameter]['2014-01-19 00:00:00'])
    df = df.resample(frequency).mean()

    # Creation of a timeseries with the positions of null values
    ts = df[parameter].isnull()

    # Check where are the intervals
    intervals = []
    in_interval = False
    end = None
    for index, value in ts.items():
        end = index.strftime('%Y-%m-%d %H:%M:%S')
        if in_interval is False and value is False:
            in_interval = True
            start = index.strftime('%Y-%m-%d %H:%M:%S')
            print(f'start: {start}')
        elif in_interval is True and value is True:
            in_interval = False
            print(f'end: {end}')
            intervals.append((start, end))
    if in_interval is True:
        intervals.append((start, end))

    return intervals



""" Implementation of WaterFrame.to_pkl(path) """
import pickle


def to_pkl(self, path_pkl=None):
    """
    Save the WaterFrame into a pickle file.

    Parameters
    ----------
        path_pkl: str
            Location of the pickle file. If path_pkl is None, the path will be the metadata['id'].

    Returns
    -------
        path_pkl: str
            Location of the pickle file.
    """
    if path_pkl is None:
        path_pkl = self.metadata['id'] + '.pkl'

    pickle.dump(self.__dict__, open(path_pkl, "wb"))
    return path_pkl



""" Implementation of WaterFrame.iplot_timeseries() """
import numpy as np
import plotly.graph_objects as go


def iplot_timeseries(list_wf, parameter_to_plot):
    """
    It creates a Plotly figure with the time-series of the input parameter.

    Parameters
    ----------
        list_wf: List of WaterFrames
            List of WaterFrames to be concatenated.
        parameter_to_plot: str
            Parameters to plot.

    Returns
    -------
        figure: dict
            Plotly figure dictionary
    """
    data = []
    for wf in list_wf:
        # Extract data
        df = wf.data[parameter_to_plot].dropna().reset_index()
        df = df.groupby(['DEPTH', 'TIME'])[parameter_to_plot].mean()
        df = df.reset_index()
        df.set_index('TIME', inplace=True)
        # df.renane(columns={parameter_to_plot: wf.metadata.get('platform_code')}, inplace=True)

        data.append(
            go.Scatter(x=df.index, y=df[parameter_to_plot],
                       # fill="tozeroy",
                       name=wf.metadata.get('platform_code')))

    # Layout
    y_label = None

    try:
        title = list_wf[0].vocabulary[parameter_to_plot]['long_name']
    except KeyError:
        title = parameter_to_plot
    try:
        y_label = list_wf[0].vocabulary[parameter_to_plot]['units']
    except KeyError:
        pass

    min_value = np.nanmin(list_wf[-1].data[parameter_to_plot])
    max_value = np.nanmax(list_wf[0].data[parameter_to_plot])

    layout = {
        'title': title,
        'yaxis': {
            # 'range': [min_value, max_value],
            'title': y_label
        },
        'margin': {'l': 50, 'r': 10, 't': 45, 'b': 30}
    }

    figure = {"data": data, "layout": layout}

    return figure



import pandas as pd
import plotly
import plotly.express as px


def iplot(self, y, x='auto', color='auto', facet_col='auto', plot_type='scatter', marginal_x='rug',
          marginal_y='box', trendline='ols', filename=None):
    """
    Get a figure using plotly.express.

    Parameters
    ----------
        y: str
            Y axes of the plot. Column of self.data to plot.
        x: str
            X axes of the plot. It can be an index or a column of self.data.
            If plot_type is 'scatter' and x is 'auto', x will the 'TIME' index of self.data. 
        color: str
            Column of self.data, auto or None. Colors of the dots of the scatter.
        facet_col: None or str
            Create different columns related to a value of a parameter.
            If facet_col is not None, marginal_y will not be visible.
            If facet_col is 'auto' and plot_type is 'scatter', facet_color is 'DEPTH' index of
            self.data.
        plot_type: str
            Type of plot figure to make.
            Options: scatter, histogram
        marginal_x: str
            Additional graph on top x axes.
            Options: None, rug, histogram, violin, box.
        marginal_y: str
            Additional graph on right y axes.
            Options: None, rug, histogram, violin, box.
        trendline: str
            Activate the trendline.
            Options:
                None
                ols (https://en.wikipedia.org/wiki/Ordinary_least_squares)
                lowess (https://en.wikipedia.org/wiki/Local_regression)
        filename: str
            Path and name of the file to save the figure in html.
    Returns
    -------
        fig: plotly figure
    """
    # Get columns to plot
    columns_to_plot = []

    # Add y axes as a list to columns to plot
    if isinstance(y, str):
        columns_to_plot = [y]
    else:
        columns_to_plot = y
    
    # Add x axes to columns to plot
    if plot_type == 'scatter':
        if x == 'auto':
            x = 'TIME'
            columns_to_plot.append('TIME')
        else:
            columns_to_plot.append(x)

    # Add color to columns to plot
    if color:
        if plot_type == 'scatter':
            if color == 'auto':
                if f'{y}_QC' in self.data.keys():
                    color = f'{y}_QC'
                    columns_to_plot.append(color)
                else:
                    color = None
            else:
                columns_to_plot.append(color)
        elif plot_type == 'histogram':
            if color == 'auto':
                if 'DEPTH' in self.data.index.names:
                    color = 'DEPTH'
                    columns_to_plot.append(color)
                else:
                    color = None
            else:
                columns_to_plot.append(color)

    # Add facet_col to columns to plot
    if facet_col:
        if plot_type == 'scatter':
            if facet_col == 'auto':
                if 'DEPTH' in self.data.index.names and 'DEPTH' not in columns_to_plot:
                    facet_col = 'DEPTH'
                    columns_to_plot.append(facet_col)
                else:
                    facet_col = None
            else:
                columns_to_plot.append(facet_col)

    # Create the df
    df = self.data.copy()
    df.reset_index(inplace=True)

    # Change 'DEPTH' and '<parameter>_QC' to type str
    if 'DEPTH' in df.keys():
        df['DEPTH'] = df['DEPTH'].apply(str)
    for key in df.keys():
        if '_QC' in key:
            df[key] = df[key].apply(str)

    # Use only the selected columns
    df = df[columns_to_plot].copy()
    df.dropna(inplace=True)

    # Make sure that 'TIME' is a timestamp
    if 'TIME' in df.keys():
        df['TIME'] = pd.to_datetime(df['TIME'])

    if plot_type == 'scatter':
        fig = px.scatter(df, 
                         x=x,
                         y=y,
                         color=color,
                         marginal_x=marginal_x,
                         marginal_y=marginal_y,
                         trendline=trendline,
                         facet_col=facet_col)
    elif plot_type == 'histogram':
        fig = px.histogram(
            df, 
            x=y,
            y=y,
            color=color)

    if filename:
        plotly.offline.plot(fig, filename=filename)

    return fig



""" Implementation of wf.qc_replace() """

def qc_replace(self, parameters=None, to_replace=0, value=1, inplace=True):
    """
    Replace the values of QC from the input parameters.

    Parameters
    ----------
        parameters: None, str or list of str.
            List of parameters to change the values of QC. Ex: ['TEMP', 'PSAL'].
        to_replace: int
            QC value to replace.
        value: int
            Value to replace any values matching to_replace with.
        inplace: bool
            If inplace, makes changes inplace and returns True.
            Otherwhise, returns a new WaterFrame.
    
    Returns
    -------
        new_wf: WaterFrame
    """

    if parameters is None:
        parameters = self.parameters
    elif isinstance(parameters, str):
        parameters = [parameters]

    data = self.data.copy()

    for parameter in parameters:
        data[f'{parameter}_QC'].replace(to_replace=to_replace, value=value, inplace=True)

    if inplace:
        self.data = data
        return True
    else:
        new_wf = self.copy()
        new_wf.data = data
        return new_wf


""" Implementation of WaterFrame.drop(parameters) """

def drop(self, parameters, inplace=True):
    """
    Remove input parameters from WaterFrame.data.

    Parameters
    ----------
        parameters: str, list of str
            Parameters of WaterFrame.data.
        inplace: bool
            If True, Drop in place and return 'True', If False, return a copy of the WaterFrame
            without the input parameters.

    Returns
    -------
        new_wf: WaterFrame
    """
    keys = []
    if isinstance(parameters, str):
        keys.append(parameters)
        keys.append(parameters + '_QC')
    elif isinstance(keys, list):
        for parameter in parameters:
            keys.append(parameter)
            keys.append(parameter + '_QC')

    if inplace:
        self.data.drop(keys, axis=1, inplace=True)
        for key in keys:
            self.vocabulary.pop(key)
        return True
    else:
        new_wf = self.copy()
        new_wf.vocabulary = self.vocabulary.copy()
        new_wf.metadata = self.metadata.copy()
        new_wf.data = self.data.drop(keys, axis=1)
        for key in keys:
            new_wf.vocabulary.pop(key)

        return new_wf



""" Output functions from a WaterFrame """
from .to_nc import to_nc
from .to_pkl import to_pkl
from .to_json import to_json
from .to_es import to_es
from .metadata_to_html import metadata_to_html
from .to_csv import to_csv



""" Implementation of WaterFrame.plot_bar(key, ax=None, average_time=None)"""
import datetime

def plot_timebar(self, keys, ax=None, time_interval_mean=None,  ylim='auto'):
    """
    Make a bar plot of the input keys.
    The bars are positioned at x with date/time. Their dimensions are given by height.

    Parameters
    ----------
        keys: list of str
                keys of self.data to plot.
        ax: matplotlib.axes object, optional (ax = None)
            It is used to add the plot to an input axes object.
        time_interval_mean: str, optional (time_interval_mean = None)
            It calculates an average value of a time interval. You can find
            all of the resample options here:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html
        ylim: 2-tuple/list, None, 'auto'
            Y axis limit. If ylim = 'auto'
    Returns
    -------
        ax: matplotlib.AxesSubplot
            Axes of the plot.
    """

    def format_year(x):
        return datetime.datetime.\
            strptime(x, '%Y-%m-%d %H:%M:%S').strftime('%Y')

    def format_day(x):
        return datetime.datetime.\
            strptime(x, '%Y-%m-%d %H:%M:%S').strftime('%Y-%m-%d')

    # Extract data
    df = self.data[keys].dropna().reset_index().set_index('TIME')
    df.index.rename("Date", inplace=True)

    # Delete DEPTH, we are not going to use it in this case
    if 'DEPTH' in df.keys():
        del df['DEPTH']

    # Resample data
    if time_interval_mean is None:
        pass
    else:
        df = df.resample(time_interval_mean).mean()

    # Get min and max values for ylim
    if ylim == 'auto':
        # df.min() returns a Series or DataFrame with the min of each column
        y_min = min(df.min().values)
        y_max = max(df.max().values)
        y_percent = (y_max - y_min) / 100
        ylim = (y_min - 5 * y_percent, y_max + 5 * y_percent)

    if isinstance(keys, list):
        ax = df[keys].plot.bar(ax=ax, legend=True, ylim=ylim)
    else:
        ax = df[keys].plot.bar(ax=ax, ylim=ylim)
        # Write axes
        try:
            ax.set_ylabel(self.vocabulary[keys]['units'])
        except KeyError:
            print("Warning: We don't know the units of", keys,
                    "Please, add info into self.meaning[", keys, "['units']")

        if time_interval_mean == 'A':
            ax.set_xticklabels([format_year(x.get_text())
                                for x in ax.get_xticklabels()], rotation=60)
        elif time_interval_mean == 'D':
            ax.set_xticklabels([format_day(x.get_text())
                                for x in ax.get_xticklabels()], rotation=60)
        elif time_interval_mean == 'M':
            ax.set_xticklabels([format_day(x.get_text())
                                for x in ax.get_xticklabels()], rotation=60)

    return ax



import pandas as pd
import plotly.express as px
import plotly.graph_objects as go


def iplot_scatter(self, y, x='TIME', trendline=None, marginal_x=None, marginal_y='histogram',
                  color='auto', symbol='DEPTH', range_y='auto', **kwds):
    """
    It uses plotly.express.scatter.
    Each data point is represented as a marker point, whose location is given by the x and y columns
    of self.data.

    Parameters
    ----------
        y: str
            Y axes, column or index of data.
        x: str
            X axes, column or index of data.
        trendline: str
            Options:
                None,
                'ols': Ordinary Least Squares regression line will be drawn for each
                    discrete-color/symbol group.
                'ols_np': Ordinary Least Squares regression line will be drawn for each
                    discrete-color/symbol group.
                    Line generated with numpy.polyfit with all values of y axes.
                'lowess': Locally Weighted Scatterplot Smoothing line will be drawn for each
                    discrete-color/symbol group.
        marginal_x: str
            If set, a horizontal subplot is drawn above the main plot, visualizing the
            x-distribution.
            Options: None, 'rug', 'box', 'violin', or 'histogram'.
        marginal_y: str
            If set, a horizontal subplot is drawn above the main plot, visualizing the
            y-distribution.
            Options: None, 'rug', 'box', 'violin', or 'histogram'.
        color: str
            Name of a column or index of data. Values from this column are used to assign color to
            marks. If color = 'auto', color = QC column of y.
        symbol: str
            Name of a column or index of data. Values from this column are used to assign symbols to
            marks.
        range_y: list
            [min value, max value] of y axes. If range_y = 'auto', range is generated between the
            mina nd max values of y axes +- 5%.
        **kwds: keywords
            plotly express scatter keywords.

    Returns
    -------
        fig: plotly.graph_objects.Figure
    """
    df = pd.DataFrame()
    _df = self.data.reset_index()

    df[y] = _df[y]
    df[x] = _df[x]

    # Range Y calculation
    y_min = min(df[y].values)
    y_max = max(df[y].values)
    y_percent = (y_max - y_min) / 100

    if range_y == 'auto':
        range_y = [y_min - 5* y_percent, y_max + 5* y_percent]

    if color == 'auto':
        color = f'{y}_QC'
        df[color] = _df[color].astype('str')
    elif color:
        df[color] = _df[color]
    
    if symbol == 'DEPTH':
        df[symbol] = _df[symbol].astype('str')
    elif symbol:
        df[symbol] = _df[symbol]

    # Save memory
    del _df

    # Dropna
    df.dropna(inplace=True)

    # Sort index TIME
    df.set_index('TIME', inplace=True)
    df.sort_index(inplace=True)
    df.reset_index(inplace=True)

    fig = px.scatter(
        df[100:150], x=x, y=y, trendline=trendline, marginal_x=marginal_x, marginal_y=marginal_y,
        color=color, symbol=symbol, range_y=range_y, **kwds)

    return fig



""" Implementation of WaterFrame.info_metadata(keys=None) """


def info_metadata(self, keys=None):
    """
    It returns a formatted string with the metadata information.

    Parameters
    ----------
        keys: string or list of strings (optional)
            Keys of WaterFrame.metadata

    Returns
    -------
        message: string
            Message with the metadata information.
    """

    if keys is None:
        keys = self.metadata.keys()

    message = ""
    for key, value in self.metadata.items():
        if key in keys and str(value).strip() != "":
            message += "  - {}: {}\n".format(key, value)

    return message[:-1]



""" Implementation of WaterFrame.iplot_timeseries() """
import numpy as np
import plotly.graph_objects as go


def iplot_timeseries(self, parameters_to_plot=None, mode='lines', marker_line_width=0,
                     marker_size=4, marker_color=None, marker_colorscale='Viridis'):
    """
    It creates a Plotly figure with the time-series of the input parameters.

    Parameters
    ----------
        parameters_to_plot: str or list (optional, parameters_to_plot=None)
            Parameters to plot.
        mode: str (optional, mode='lines')
            Plotting points (makers), lines or lines and points (lines+markers).
        marker_line_width: float (optional, marker_line_width=0)
            Width of the line arround the marker.
        marker_size: float (optional, marker_size=4)
            Size of the marker.
        marker_color: str (optional, marker_color=None)
            Parameter name. Change the color of the marker related to a parameter.
        marker_colorscale: str (optional, marker_colorscale='Viridis')
            Color scale of markers. Options: Blackbody, Bluered, Blues, Earth, Electric, Greens,
            Greys, Hot, Jet, Picnic, Portland, Rainbow, RdBu, Reds, Viridis, YlGnBu, YlOrRd
    Returns
    -------
        figure: dict
            Plotly figure dictionary
    """

    # Show color scale
    showscale = False

    if parameters_to_plot is None:
        parameters_to_plot = self.parameters
    elif isinstance(parameters_to_plot, str):
        parameters_to_plot = [parameters_to_plot]

    parameters_and_colors = parameters_to_plot.copy()
    if marker_color is not None:
        parameters_and_colors.append(marker_color)

    # Extract data
    df = self.data[parameters_and_colors].dropna().reset_index()
    df = df.groupby(['DEPTH', 'TIME'])[parameters_and_colors].mean()
    df.reset_index(inplace=True)
    df.set_index('TIME', inplace=True)

    if marker_color is not None:
        marker_color = df[marker_color]
        showscale = True

    data = [go.Scattergl(
        x=df.index,
        y=df[parameter],
        fill="tozeroy",
        mode=mode,
        marker_line_width=marker_line_width,
        marker_size=marker_size,
        marker_color=marker_color,
        marker_showscale=showscale,
        marker_colorscale=marker_colorscale,
        name=parameter)
            for parameter in parameters_to_plot]

    # Layout
    y_label = None
    if len(parameters_to_plot) == 1:
        try:
            title = self.vocabulary[parameters_to_plot[0]]['long_name']
        except KeyError:
            title = parameters_to_plot[0]
        try:
            y_label = self.vocabulary[parameters_to_plot[0]]['units']
        except KeyError:
            pass
    else:
        title = "_".join(parameters_to_plot)
    min_value = None
    max_value = None
    for parameter in parameters_to_plot:

        _min = np.nanmin(self.data[parameters_to_plot])
        _max = np.nanmax(self.data[parameters_to_plot])

        if min_value is None or _min < min_value:
            min_value = _min

        if max_value is None or _max > max_value:
            max_value = _max

    layout = {
        'title': title,
        'yaxis': {
            'range': [min_value, max_value],
            'title': y_label
        },
        'margin': {'l': 50, 'r': 10, 't': 45, 'b': 30}
    }

    figure = {"data": data, "layout": layout}

    return figure



""" Implementation of WaterFrame.plot(**kwds)"""


def plot(self, **kwds):
    """
    It calls the pandas DataFrame.plot() method.

    Parameters
    ----------
        **kwds: arguments
            pandas plot() arguments.

    Returns
    -------
        axes : matplotlib.axes.Axes or numpy.ndarray of them
    """
    axes = self.data.plot(**kwds)
    return axes



from elasticsearch import Elasticsearch


def es_create_indexes(data_index_name='data', metadata_index_name='metadata',
                   summary_index_name='summary', delete_previous_indexes=True, **kwargs):
    """
    Creation of ElasticSearch Indexes to save a WaterFrame object.
    
    Parameters
    ----------
        data_index_name: str
            Name of the ElasticSearch index that contains the WaterFrame.data documents.
        metadata_index_name: str
            Name of the ElasticSearch index that contains the WaterFrame.metadata documents.
        summary_index_name: str
            Name of the ElasticSearch index that contains the summary documentation.
        delete_previous_indexes: bool
            Delete previous indexes with the input names and all their documents.
        **kwargs: Elasticsearch object creation arguments.
            See https://elasticsearch-py.readthedocs.io/en/master/index.html#
    
    Returns
    -------
        success: bool
            If sucess is true, indexes where created.
    """
    def create_data_index():
        # Index settings
        settings = {
            'mappings': {
                'properties': {
                    'parameter': {'type': 'keyword'},
                    'time': {
                        'type': 'date',
                        'format': 'yyyy-MM-dd HH:mm:ss'
                    },
                    'time_qc': {'type': 'integer'},
                    'depth': {'type': 'float'},
                    'depth_qc': {'type': 'integer'},
                    'value': {'type': 'float'},
                    'value_qc': {'type': 'integer'},
                    'metadata_id': {'type': 'keyword'},
                }
            }
        }

        es.indices.create(index=data_index_name, body=settings)

    def create_metadata_index():

        # Index settings
        settings = {
            'mappings': {
                'properties': {
                    'platform_code': {'type': 'keyword'},
                    'platform_name': {'type': 'keyword'},
                    'ices_platform_code': {'type': 'keyword'},
                    'institution': {'type': 'text', 'analyzer': 'standard'},
                    'date_update': {'type': 'date'},
                    'site_code': {'type': 'keyword'},
                    'wmo_platform_code': {'type': 'keyword'},
                    'source': {'type': 'text', 'analyzer': 'standard'},
                    'history': {'type': 'text', 'analyzer': 'standard'},
                    'data_mode': {'type': 'keyword'},
                    'references': {'type': 'text', 'analyzer': 'standard'},
                    'comment': {'type': 'text', 'analyzer': 'standard'},
                    'summary': {'type': 'text', 'analyzer': 'standard'},
                    'id': {'type': 'keyword'},
                    'area': {'type': 'keyword'},
                    'geospatial_min': {'type': 'geo_point'},
                    'geospatial_max': {'type': 'geo_point'},
                    'geospatial_vertical_min': {'type': 'float'},
                    'time_coverage_start': {'type': 'date'},
                    'time_coverage_end': {'type': 'date'},
                    'institution_references': {'type': 'text', 'analyzer': 'standard'},
                    'contact': {'type': 'text', 'analyzer': 'standard'},
                    'author': {'type': 'text', 'analyzer': 'standard'},
                    'pi_name': {'type': 'text', 'analyzer': 'standard'},
                    'update_interval': {'type': 'keyword'},
                    'wmo_inst_type': {'type': 'keyword'},
                    'qc_manual': {'type': 'text', 'analyzer': 'standard'},
                    'data_type': {'type': 'text', 'analyzer': 'standard'},
                    'naming_authority': {'type': 'keyword'},
                    'quality_control_indicator': {'type': 'integer'},
                    'distribution_statement': {'type': 'text', 'analyzer': 'standard'},
                    'quality_index': {'type': 'keyword'},
                    'citation': {'type': 'text', 'analyzer': 'standard'},
                    'format_version': {'type': 'float'},
                    'Conventions': {'type': 'text', 'analyzer': 'standard'},
                    'title': {'type': 'text', 'analyzer': 'standard'},
                    'data_assembly_center': {'type': 'keyword'},
                    'last_location_observation': {'type': 'geo_point'},
                    'last_date_observation': {'type': 'date'},
                    'parameters': {'type': 'keyword'},
                    'site': {'type': 'keyword'},
                    'network': {'type': 'keyword'},
                }
            }
        }

        es.indices.create(index=metadata_index_name, body=settings)

    def create_summary_index():
        # Index settings
        settings = {
            'mappings': {
                'properties': {
                    'parameters': {
                        'properties': {
                            'long_name': {'type': 'keyword'},
                            'acronym': {'type': 'keyword'}
                        },
                    },
                    'networks': {'type': 'keyword'},
                    'sites': {'type': 'keyword'},
                }
            }
        }

        # Creation of the index
        es.indices.create(index=summary_index_name, body=settings)

    es = Elasticsearch(**kwargs)

    if delete_previous_indexes:
        # Delete indexes, ignore if indexes do not exist
        es.indices.delete(index=data_index_name, ignore=404)
        es.indices.delete(index=metadata_index_name, ignore=404)
        es.indices.delete(index=summary_index_name, ignore=404)
    
    # Index creation
    create_data_index()
    create_metadata_index()
    create_summary_index()

    success = True
    return success



""" Setup config to install mooda """

import os
from distutils.core import setup
from setuptools import find_packages


NAME = 'mooda'
VERSION = '1.5.0'
DESCRIPTION = 'Module for Ocean Observatory Data Analysis'
LONG_DESCRIPTION = ("""
MOODA - Module for Ocean Observatory Data Analysis

Mooda is a python package designed mainly for oceanographers and marine science students. It is"""
                    """based on a power scripting system for:

* open and analyze data files from scientific instrumentation and platforms
* generate data quality control
* make plots that are commonly used in the oceanographic community
* make data files in netCDF and CSV format

Check the documentation on https://github.com/rbardaji/mooda.
""")
CLASSIFIERS = ['Development Status :: 3 - Alpha',
               'Environment :: Console',
               'Natural Language :: English',
               'License :: OSI Approved :: MIT License',
               'Programming Language :: Python :: 3.7',
               'Topic :: Software Development :: Libraries :: Python Modules',
               'Topic :: Scientific/Engineering :: Physics',
               'Intended Audience :: Education',
               'Intended Audience :: Science/Research',
               'Intended Audience :: Developers']
KEYWORDS = ['ocean',
            'sea',
            'EMODnet',
            'EMSO']
URL = 'https://github.com/rbardaji/mooda'
AUTHOR = 'Raul Bardaji Benach'
AUTHOR_EMAIL = 'rbardaji@gmail.com'
LICENSE = 'MIT'
PACKAGES = find_packages()
REQUIREMENTS = ['requirements.txt']
INSTALL_REQUIRES = sorted(
    set(
        line.partition('#')[0].strip()
        for file in (os.path.join(os.path.dirname(__file__), file)
                     for file in REQUIREMENTS)
        for line in open(file)) - {''})

setup(name=NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      classifiers=CLASSIFIERS,
      keywords=KEYWORDS,
      url=URL,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      packages=PACKAGES,
      install_requires=INSTALL_REQUIRES,
      include_package_data=True,
      zip_safe=False)



""" Implementation fo WaterFrame.to_nc() """


def to_csv(self, path=None):
    """
    Create a CSV file with the WaterFrame data.
    The metadata and vocabulary will be placed in the first lines of the file with a # as first
    character of the line.

    Parameters
    ----------
        path: str
            Location and filename of the csv file.
            If path is None, the filename will be the
            metadata['id'].
    
    Returns
    -------
        path: str
            Location and filename of the csv file.
    """

    filename = path
    if path is None:
        filename = self.metadata['id'] + '.csv'

    with open(filename, 'w') as f:
        f.write('# METADATA\n')
        for key, value in self.metadata.items():
            if value != '':
               f.write(f'# {key}: {value}\n')
        f.write('\n')
        f.write('# VOCABULARY\n')
        for key, definition in self.vocabulary.items():
            f.write(f'# {key}:\n')
            for key_definition, value_definition in definition.items():
               f.write(f'#   - {key_definition}: {value_definition}\n')
        f.write('\n')

    # Save data in a csv
    self.data.to_csv(filename, mode='a')

    return filename



""" Implementation of a function to read NetCDF files from Ocean Observatory Iniciative (OOI) """
import xarray as xr
from ..waterframe import WaterFrame


def read_nc_ooi(path, clean_data=True):
    """
    Open a NetCDF file from OOI. This function was tested with CTD files.
    Web - https://oceanobservatories.org

    Parameters
    ----------
        path: str
            Path of the NetCDF file.
        clean_data: bool
            It deletes variables with DC in name.

    Returns
    -------
        wf: WaterFrame
    """

    def add_pres(key_in):
        """ Add parameter PRES """
        # Data
        wf.data.rename(columns={key_in: 'PRES'}, inplace=True)
        wf.data['PRES_QC'] = 0

        # vocabulary
        wf.vocabulary['PRES'] = {
            "standard_name": "sea_water_pressure",
            "long_name": "Sea pressure",
            "units": "dbar",
            "valid_min": "0.0",
            "valid_max": "12000.0"
        }

    def add_temp(key_in):
        """ Add parameter TEMP """
        # Data
        wf.data.rename(columns={key_in: 'TEMP'}, inplace=True)
        wf.data['TEMP_QC'] = 0

        # vocabulary
        wf.vocabulary['TEMP'] = {
            "valid_min": "-5.0",
            "valid_max": "35.0",
            "ancillary_variables": "TEMP_QC",
            "standard_name": "sea_water_temperature",
            "units": "degrees_C",
            "long_name": "Sea temperature"
        }

    def add_cndc(key_in):
        """ Add parameter CNDC """
        # Data
        wf.data.rename(columns={key_in: 'CNDC'}, inplace=True)
        wf.data['CNDC_QC'] = 0

        # vocabulary
        wf.vocabulary['CNDC'] = {
            "valid_min": "0.0",
            "valid_max": "7.0",
            "ancillary_variables": "CNDC_QC",
            "standard_name": "sea_water_electrical_conductivity",
            "units": "S m-1",
            "long_name": "Electrical conductivity"
        }

    def add_psal(key_in):
        """ Add parameter PSAL """
        # Data
        wf.data.rename(columns={key_in: 'PSAL'}, inplace=True)
        wf.data['PSAL_QC'] = 0

        # vocabulary
        wf.vocabulary['PSAL'] = {
            "valid_min": "0.0",
            "valid_max": "42.0",
            "ancillary_variables": "PSAL_QC",
            "standard_name": "sea_water_practical_salinity",
            "units": "PSU",
            "long_name": "Practical salinity"
        }

    # WaterFrame creation
    wf = WaterFrame(path)

    # Set index
    try:
        wf.data.rename(columns={'time': 'TIME'}, inplace=True)
        # Set index to TIME
        wf.data.set_index('TIME', inplace=True)
    except KeyError:
        pass

    keys = wf.data.keys()
    # OOI sometimes contains "ctmo", other no... extrange things.

    if any("ctdmo" in key_str for key_str in keys):
        for key in keys:

            # Only use columns with "ctmo" in the name
            # We are going to assign QC to 0, CHECK IT
            if key == "ctdmo_seawater_pressure":
                add_pres(key)
            elif key == "ctdmo_seawater_temperature":
                add_temp(key)
            elif key == "ctdmo_seawater_conductivity":
                add_cndc(key)
            else:
                del wf.data[key]
    elif any("ctdpf" in key_str for key_str in keys):
        for key in keys:

            # Only use columns with "ctmo" in the name
            # We are going to assign QC to 0, CHECK IT
            if key == "ctdpf_ckl_seawater_pressure":
                add_pres(key)
            elif key == "ctdpf_ckl_seawater_temperature":
                add_temp(key)
            elif key == "ctdpf_ckl_seawater_conductivity":
                add_cndc(key)
            else:
                del wf.data[key]
    else:
        for key in keys:
            if key == "pressure":
                add_pres(key)
            elif key == "temp":
                add_temp(key)
            elif key == "conductivity":
                add_cndc(key)
            elif key == "practical_salinity":
                add_psal(key)

    # Adding platform_code to metadata
    wf.metadata['platform_code'] = wf.metadata['node']
    return wf



import json
from elasticsearch import Elasticsearch, exceptions
from time import sleep


def to_es(self, data_index_name='data', metadata_index_name='metadata',
          summary_index_name='summary', qc_to_ingest=[0, 1], **kwargs):
    """
    Injestion of the WaterFrame into a ElasticSeach DB.

    Parameters
    ----------
        data_index_name: str
            Name of the ElasticSearch index that contains the WaterFrame.data documents.
        metadata_index_name: str
            Name of the ElasticSearch index that contains the WaterFrame.metadata documents.
        summary_index_name: str
            Name of the ElasticSearch index that contains the summary documents.
        qc_to_intest: list of int
            QC Flags of data to be ingested to the ElasticSearch DB.
        **kwargs: Elasticsearch object creation arguments.
            See https://elasticsearch-py.readthedocs.io/en/master/index.html#
    """
    def data_ingestion(es, data_index_name, data_dict):

        data_id = f"{data_dict['metadata_id']}_{data_dict['parameter']}_{data_dict['time'].replace(' ', 'T')}_{data_dict['depth']}"
        data_json = json.dumps(data_dict)
        es.index(index=data_index_name, id=data_id, body=data_json)

        return True

    def summary_ingestion(es, summary_index_name, summary_dict):

        try:
            # Read emso-stats
            response = es.get(index='emso-stats', id='datalab')
            es_summary = response['_source']

            es_summary['sites'] = list(set(es_summary['sites'] + summary_dict['sites']))
            es_summary['networks'] = list(set(es_summary['networks'] + summary_dict['networks']))
            param_list = []

            for stats_parameter in summary_dict['parameters']:

                tengui = False
                for summary_parameter in es_summary['parameters']:
                    if summary_parameter['acronym'] == stats_parameter['acronym'] or \
                        'QC' in summary_parameter['acronym']:
                        tengui = True
                        break
                if not tengui:
                    param_list.append(stats_parameter)
            es_summary['parameters'] += param_list

            # Ingest updated stats
            es.index(index=summary_index_name, id='datalab', body=es_summary)
            return True

        except exceptions.NotFoundError:
            summary_json = json.dumps(summary_dict)
            es.index(index=summary_index_name, id='datalab', body=summary_json)
            return True

    def get_metadata(es, metadata_index_name, search):
        """
        Get a list with the metadata information.
        """

        search_body = {
            'query': {
                'bool': {
                    'must': []
                }
            }  
        }
        if search.get('platform_code'):
            search_body['query']['bool']['must'].append(
                {'term': {'platform_code': search.get('platform_code')}})
        if search.get('geospatial_lat_min'):
            search_body['query']['bool']['must'].append(
                {'term': {'geospatial_lat_min': search.get('geospatial_lat_min')}})
        if search.get('geospatial_lon_min'):
            search_body['query']['bool']['must'].append(
                {'term': {'geospatial_lon_min': search.get('geospatial_lon_min')}})

        # Search query in metadata index
        response = es.search(index=metadata_index_name, body=search_body)
        
        if response['hits']['hits']:
            # Create a metadata dict with the metadata info and the id of the index
            for hit in response['hits']['hits']:
                return hit['_id'], hit['_source']['time_coverage_start'], hit['_source']['time_coverage_end']
        else:
            return False

    def metadata_ingestion(es, metadata_index_name, metadata_dict):

        metadata_search = {
            'platform_code': metadata_dict['platform_code'],
            'time_coverage_start': metadata_dict['time_coverage_start'],
            'time_coverage_end': metadata_dict['time_coverage_end']
        }

        result = get_metadata(es, metadata_index_name, metadata_search)

        if result:
            metadata_id = result[0]
            start = result[1]
            end = result[2]

            metadata_dict['id'] = metadata_id
            if metadata_dict.get('time_coverage_start') > start:
                metadata_dict['time_coverage_start'] = start
            if metadata_dict.get('time_coverage_end') < end:
                metadata_dict['time_coverage_end'] = end

        metadata_json = json.dumps(metadata_dict)
        # Create or update document
        es.index(index=metadata_index_name, id=metadata_dict['id'], body=metadata_json)
        return True
    
    es = Elasticsearch(**kwargs)

    # Add parameters in metadata information
    self.metadata['parameters'] = self.parameters

    ok = metadata_ingestion(es, metadata_index_name, self.metadata)
    if ok:
        print(f"Metadata {self.metadata['id']} ingested")
    else:
        print(f"Error: Metadata {self.metadata['id']} not ingested")
    
    summary_dict = {
        'sites': [self.metadata['site_code']],
        'networks': [self.metadata['network']],
        'parameters': []}
    for key, value in self.vocabulary.items():
        if '_QC' not in key or 'TIME' not in key or 'DEPH' not in key:
            try:
                summary_dict['parameters'].append({'long_name': value['long_name'], 'acronym': key})
            except KeyError:
                print(f"Caution: {key} doesn't contain 'long_name'")

    ok = summary_ingestion(es, summary_index_name, summary_dict)
    if ok:
        print('Stats updated')
    else:
        print('Error: Stats not updated')

    for parameter in self.parameters:
        mini_wf = self.copy()
        try:
            mini_wf.data = self.data[[parameter, parameter+'_QC', 'DEPH', 'DEPH_QC', 'TIME_QC']]
        except KeyError:
            mini_wf.data = self.data[[parameter, parameter+'_QC', 'TIME_QC']]
            mini_wf.data['DEPH'] = mini_wf.data.index.get_level_values('DEPTH')
            if 'DEPTH_QC' in self.data.keys():
                mini_wf.data['DEPH_QC'] = self.data['DEPTH_QC']
            else:
                mini_wf.data['DEPH_QC'] = 0

        mini_wf.data = mini_wf.data.dropna()

        mini_wf.data.reset_index(inplace=True)
        mini_wf.data.set_index('TIME', inplace=True)

        df_init = mini_wf.data

        for qc_value in qc_to_ingest:
            df = df_init[df_init[parameter+'_QC'] == qc_value]

            for num, (index, row) in enumerate(df.iterrows()):
                data_dict = {
                    'parameter': parameter,
                    'time': str(index)[:19],
                    'time_qc': int(row['TIME_QC']),
                    'depth': str(row['DEPH']),
                    'depth_qc': int(row['DEPH_QC']),
                    'value': str(row[parameter]),
                    'value_qc': int(row[parameter+'_QC']),
                    'metadata_id': self.metadata['id']
                }

                ok = data_ingestion(es, data_index_name, data_dict)
                if ok:
                    print(f"{data_dict['parameter']} from {data_dict['metadata_id']} ingested: {num} of {len(df.index)}")
                else:
                    print(f"ERROR: {data_dict['parameter']} from {data_dict['metadata_id']} NOT ingested: {num} of {len(df.index)}")

                # Wait for ES operations
                sleep(0.01)
